open Common;;
open Types;;
open Equations;;
open Typing_tree;;
open Proj;;

type subst = SSimple of typ_var * typ * terr
           | SDupl of int * typ_var list;;

let print_subst = function
   SSimple (t, tau, terr) -> print_string "Substituting: "; print_typ_var t; print_string " => "; print_typ tau;
                             print_string "   "; print_terr terr; print_newline ()
 | SDupl (n, dom) -> print_string "Duplicating: "; print_string "dupl("; print_int n; print_string ",";
                     comma_list "" print_typ_var "," dom; print_string ")"; print_newline ();;

let print_subst_list sbs = List.iter print_subst (List.rev sbs);;

let simple_subst_typ t tau =
 let rec loop = function
    TypVar t' when t = t' -> tau
  | TypVar t' -> TypVar t'
  | TypFun (ts, tau) -> if List.mem t ts then failwith "simple_subst_typ-loop";
                        TypFun (ts, loop tau)
 in loop;;

let simple_subst_terr t terr terr' =
 if List.mem t terr' then union terr (remove1 t terr') else terr';;

let simple_subst_eq (t, tau, terr) (tau', t', sigma', terr') =
 if t = t' then failwith "simple_subst_eq";
 (simple_subst_typ t tau tau', t', simple_subst_typ t tau sigma', simple_subst_terr t terr terr');;

let dupl_subst_pattern n dom ts =
 let foo x = if List.mem x dom then duplicate n proj_typ_var x else [x] in
 List.concat (List.map foo ts);;

let dupl_subst_typ n dom =
 let rec loop = function
    TypVar t -> if List.mem t dom then failwith "dupl_subst_typ-loop" else TypVar t
  | TypFun (ts, tau) -> TypFun (dupl_subst_pattern n dom ts, loop tau)
 in loop;;

let dupl_subst_terr n dom terr =
 List.fold_right (fun t' -> simple_subst_terr t' (duplicate n proj_typ_var t')) dom terr;;

let dupl_subst_eq (n, dom) (tau, t, sigma, terr) =
 if List.mem t dom
 then duplicate n proj_eq (tau, t, sigma, terr)
 else [dupl_subst_typ n dom tau, t, dupl_subst_typ n dom sigma, dupl_subst_terr n dom terr];;

let subst_eq sbs eq = match sbs with
   SSimple (t, tau, terr) -> [simple_subst_eq (t, tau, terr) eq]
 | SDupl (n, dom) -> dupl_subst_eq (n, dom) eq;;

type subst2 = SSimple2 of typ_var * typ2
            | SDupl2 of int * typ_var list;;

let subst_to_subst2 = function
   SSimple (t, tau, _) -> SSimple2 (t, typ_to_typ2 tau)
 | SDupl (n, ts) -> SDupl2 (n, ts);;

let print_subst2 = function
   SSimple2 (t, tau) -> print_typ_var t; print_string " => "; print_typ2 tau
 | SDupl2 (n, dom) -> print_string "dupl("; print_int n; print_string ",";
                      comma_list "" print_typ_var "," dom; print_string ")";;

let simple_subst_typ2 (t, tau) =
 let rec loop = function
    TypVar2 t' when t = t' -> tau
  | TypVar2 t' -> TypVar2 t'
  | TypFun2 (sigmas, sigma') -> TypFun2 (List.map loop sigmas, loop sigma')
 in loop;;

let dupl_subst_typ2 n dom =
 let rec loop2 tau =
  if subset (fv_typ2 tau) dom
  then duplicate n proj_typ2 tau
  else [loop tau]
 and loop = function
    TypVar2 t -> if List.mem t dom then failwith "dupl_subst_typ2-loop" else TypVar2 t
  | TypFun2 (sigmas, sigma') -> TypFun2 (List.concat (List.map loop2 sigmas), loop sigma')
 in loop;;

let subst_typ2 sbs tau = match sbs with
   SSimple2 (t, sigma) -> simple_subst_typ2 (t, sigma) tau
 | SDupl2 (n, dom) -> dupl_subst_typ2 n dom tau;;

let simple_subst_typing_tree s = map_typing_tree (simple_subst_typ2 s);;

let dupl_subst_binding n dom (x, tau) =
 if subset (fv_typ2 tau) dom
 then List.map (fun sigma -> (x, sigma)) (duplicate n proj_typ2 tau)
 else [(x, dupl_subst_typ2 n dom tau)];;

let dupl_subst_env n dom env = List.concat (List.map (dupl_subst_binding n dom) env);;

let dupl_subst_typing n dom (env, m, tau) = (dupl_subst_env n dom env, m, dupl_subst_typ2 n dom tau);;

let dupl_subst_typing_tree nb dom =
 let rec loop2 t =
  if subset (fv_typing_tree t) dom
  then duplicate nb proj_typing_tree t
  else [loop t]
 and loop = function
    TTId typing -> TTId (dupl_subst_typing nb dom typing)
  | TTAppl (m, ns, typing) -> TTAppl (loop m, List.concat (List.map loop2 ns), dupl_subst_typing nb dom typing)
  | TTLambda (m, typing) -> TTLambda (loop m, dupl_subst_typing nb dom typing)
  | TTForget (m, n, typing) -> TTForget (loop m, loop n, dupl_subst_typing nb dom typing)
 in loop;;

let subst_typing_tree sbs ttree = match sbs with
   SSimple2 (t, tau) -> simple_subst_typing_tree (t, tau) ttree
 | SDupl2 (n, dom) -> dupl_subst_typing_tree n dom ttree;;

let simple_subst_eq2 (t, tau) (sigma, t') =
 if t = t' then failwith "simple_subst_eq2";
 (simple_subst_typ2 (t, tau) sigma, t');;

let subst_eq2 sbs eq = match sbs with
   SSimple2 (t, tau) -> simple_subst_eq2 (t, tau) eq
 | SDupl2 _ -> failwith "subst_eq2";;
