open Common;;
open Types;;

type ident_var = string;;

let print_ident_var = print_string;;

type term =
   TermId of ident_var
 | TermAppl of term * term
 | TermLambda of ident_var * term
 | TermForget of term * term;;

let print_term =
 let rec loop par = function
    TermId x -> print_ident_var x
  | TermAppl (m, n) -> if par then print_string "(";
                       loop true m; print_string " "; loop true n;
                       if par then print_string ")"
  | TermLambda (x, m) -> if par then print_string "(";
                         print_string "\\"; print_ident_var x; print_string " "; loop false m;
                         if par then print_string ")"
  | TermForget (m, n) -> print_string "["; loop false m; print_string ", "; loop false n; print_string "]"
 in loop false;;

type typed_term =
   TypedTermId of ident_var * typ_var
 | TypedTermAppl of typed_term * typed_term * typ_var
 | TypedTermLambda of ident_var * typed_term * typ
 | TypedTermForget of typed_term * typed_term * typ;;

let rec fv_typed_term = function
   TypedTermId (_, t) -> [t]
 | TypedTermAppl (tm, tn, t) -> union [t] (union (fv_typed_term tm) (fv_typed_term tn))
 | TypedTermLambda (_, tm, tau) -> union (fv_typ tau) (fv_typed_term tm)
 | TypedTermForget (tm, tn, tau) -> union (fv_typ tau) (union (fv_typed_term tm) (fv_typed_term tn));;

let get_type = function
   TypedTermId (_, t) -> TypVar t
 | TypedTermAppl (_, _, t) -> TypVar t
 | TypedTermLambda (_, _, t) -> t
 | TypedTermForget (_, _, t) -> t;;

let rec var_typ x = function
   TypedTermId (y, t) -> if x = y then [t] else []
 | TypedTermAppl (tm, tn, _) -> union (var_typ x tm) (var_typ x tn)
 | TypedTermLambda (y, tm, _) -> if x = y then [] else var_typ x tm
 | TypedTermForget (tm, tn, _) -> union (var_typ x tm) (var_typ x tn);;

let rec term_to_typed_term = function
   TermId x -> TypedTermId (x, fresh_typ_var ())
 | TermAppl (m, n) -> let tm = term_to_typed_term m in
                      let tn = term_to_typed_term n in
                      TypedTermAppl (tm, tn, fresh_typ_var ())
 | TermLambda (x, m) -> let tm = term_to_typed_term m in
                        TypedTermLambda (x, tm, TypFun (var_typ x tm, get_type tm))
 | TermForget (m, n) -> let tm = term_to_typed_term m in
                        let tn = term_to_typed_term n in
                        TypedTermForget (tm, tn, get_type tm);;

let rec typed_term_to_term = function
   TypedTermId (x, _) -> TermId x
 | TypedTermAppl (tm, tn, _) -> TermAppl (typed_term_to_term tm, typed_term_to_term tn)
 | TypedTermLambda (x, tm, _) -> TermLambda (x, typed_term_to_term tm)
 | TypedTermForget (tm, tn, _) -> TermForget (typed_term_to_term tm, typed_term_to_term tn);;

let rec print_typed_term = function
   TypedTermId (x, t) -> print_ident_var x; print_string ":"; print_typ_var t
 | TypedTermAppl (tm, tn, t) -> print_string "("; print_typed_term tm; print_string " "; print_typed_term tn;
                                print_string "):"; print_typ_var t
 | TypedTermLambda (x, tm, t) -> print_string "(\\"; print_ident_var x; print_string " "; print_typed_term tm;
                                 print_string "):"; print_typ t
 | TypedTermForget (tm, tn, t) -> print_string "["; print_typed_term tm; print_string ","; print_typed_term tn;
                                  print_string "]:"; print_typ t;;
